class User:
    __slots__ = ("id", "name", "exp", "level", "weeklyexp")

    def __init__(self, data):
        self.set_data(data)

    def __str__(self) -> str:
        return self.name

    def set_data(self, data):
        self.id = int(data["id"])
        self.name = data["username"]
        self.exp = int(data["exp"])
        self.level = int(data["level"])
        self.weeklyexp = int(data["weeklyExp"])


class Leaderboard:
    __slots__ = ("id", "usercount", "leaderboard")

    def __init__(self, data):
        self.set_data(data)

    def __str__(self) -> str:
        return str(self.id)

    def set_data(self, data):
        self.id = int(data["id"])
        self.usercount = int(data["total_count"])
        self.leaderboard = data["data"]


class Rewards:
    __slots__ = ("id", "count", "roles")

    def __init__(self, data):
        self.set_data(data)

    def __str__(self) -> str:
        return str(self.id)

    def set_data(self, data):
        self.id = int(data["id"])
        self.count = int(data["count"])
        self.roles = [(role["level"], role["roleID"]) for role in data["data"]]



import logging
from typing import Dict

import asyncio
import aiohttp

from datetime import datetime

from .exceptions import NotFound, InvalidToken, Ratelimited, AmariServerError

from .objects import User, Leaderboard, Rewards

BASE_URL = "https://amaribot.com/api/v1/"

logger = logging.getLogger(__name__)

HTTP_response_errors = {
    404: NotFound,
    403: InvalidToken,
    429: Ratelimited,
    500: AmariServerError,
}


def check_response_for_errors(resp_status):
    if resp_status in HTTP_response_errors:
        raise HTTP_response_errors[resp_status]


class AmariBot:
    def __init__(self, token: str):

        if not isinstance(token, str):
            raise TypeError("The token must be a string.")

        self.baseurl = BASE_URL

        self.default_headers = {"Authorization": token}

        # Anti Ratelimit section
        self.requests = []

        self.max_requests = 60
        # Value in seconds
        self.request_period = 60

    def update_ratelimit(self):
        for request_time in self.requests:
            if (datetime.utcnow() - request_time).seconds >= self.request_period:
                self.requests.remove(request_time)

    async def check_and_update_ratelimit(self):
        self.update_ratelimit()

        if len(self.requests) == self.max_requests - 1:
            await self.wait_for_ratelimit_end()

    async def wait_for_ratelimit_end(self):
        test = 0
        for count in range(1, 6):
            wait_amount = 2 ** count
            test += wait_amount

            logger.warning(
                f"Slow down, you are about to be rate limited. Trying again in {wait_amount} seconds."
            )
            await asyncio.sleep(wait_amount)

            self.update_ratelimit()
            if len(self.requests) != self.max_requests - 1:
                break

    async def get_user(self, guild_id: int, user_id: int) -> User:
        if not isinstance(guild_id, int):
            raise TypeError("The guild_id must be an int.")

        if not isinstance(user_id, int):
            raise TypeError("The user_id must be an int.")

        data = await self.create_request(
            "guild/{guild_id}/member/{user_id}".format(
                guild_id=guild_id, user_id=user_id
            ),
        )

        return User(data)

    async def get_leaderboard(
        self, guild_id: int, *, weekly: bool = False, page: int = 1, limit: int = 50
    ) -> Leaderboard:
        if not isinstance(guild_id, int):
            raise TypeError("The guild_id must be an int.")

        if not isinstance(weekly, bool):
            raise TypeError("The weekly variable must be a bool.")

        if not isinstance(page, int):
            raise TypeError("The page must be an int.")

        if not isinstance(limit, int):
            raise TypeError("The limit must be an int.")

        params = {"page": page, "limit": limit}

        data = await self.create_request(
            "guild/{lbtype}/{guild_id}".format(
                lbtype="weekly" if weekly else "leaderboard", guild_id=guild_id
            ),
            params=params,
        )

        data["id"] = guild_id
        return Leaderboard(data)

    async def get_rewards(
        self, guild_id: int, *, page: int = 1, limit: int = 50
    ) -> Rewards:
        if not isinstance(guild_id, int):
            raise TypeError("The guild_id must be an int.")

        if not isinstance(page, int):
            raise TypeError("The page must be an int.")

        if not isinstance(limit, int):
            raise TypeError("The limit must be an int.")

        params = {"page": page, "limit": limit}

        data = await self.create_request(
            "guild/rewards/{guild_id}".format(guild_id=guild_id),
            params=params,
        )

        data["id"] = guild_id
        return Rewards(data)

    async def create_request(self, endpoint: str, *, params: Dict = {}) -> Dict:
        await self.check_and_update_ratelimit()

        async with aiohttp.ClientSession() as session:
            async with session.get(
                url=self.baseurl + endpoint,
                headers=self.default_headers,
                params=params,
            ) as response:

                check_response_for_errors(response.status)

                self.requests.append(datetime.utcnow())
                return await response.json()



class BaseHTTPRequestError(Exception):
    """Base error for request responses."""


class NotFound(BaseHTTPRequestError):
    """Raised when the guild or user is not found."""

    def __init__(self):
        super().__init__("Guild or user was not found.")


class InvalidToken(BaseHTTPRequestError):
    """Raised when the authentication key is invalid."""

    def __init__(self):
        super().__init__(
            "Please enter a valid authentication key.\n"
            "You can obtain your key at https://amaribot.com/developer/yourid"
        )


class Ratelimited(BaseHTTPRequestError):
    """Raised when ratelimit responses are recieved."""

    def __init__(self, message="Slow down! You are being ratelimited!"):
        super().__init__(message)


class AmariServerError(BaseHTTPRequestError):
    """Raised when their is an internal error in the Amari servers."""

    def __init__(self):
        super().__init__("There was an internal error in the Amari servers.")



from setuptools import setup
import amari.__init__ as init

readme = ""
with open("README.rst") as file:
    readme = file.read()

setup(
    name="amari",
    version=init.__version__,
    url="https://github.com/TheF1ng3r/amari.py",
    project_urls={"Documentation": "https://amaripy.gitbook.io/docs/"},
    description="An asynchronous API wrapper for the Amari API.",
    long_description=readme,
    author="TheF1ng3r",
    license="MIT",
    install_requires="aiohttp>=3.6.0,<3.8.0",
    python_requires=">=3.5.0",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    packages=["amari"],
)
